/***************************************************************************************************
 *                                        FREDuino ver. 1.0.1                                      *
 ***************************************************************************************************                                        
 *                                                                                                 *
 * Ovladac pro modelovou zeleznici komunikujici po sbernici LocoNet (Digitrax)                     *
 *                                                                                                 *
 * Software vyuziva funkcionality Arduino knihoven LocoNet, jejichz autorem je Alex Shepherd       *
 * https://github.com/mrrwa/LocoNet/archive/master.zip                                             *
 *                                                                                                 *
 * Na nektere casti techto knihoven se mohou vztahovat autorska prava Digitrax a Uhlenbrock        *
 *                                                                                                 *
 * Zakladni funkcionalita je blizka ovladacum FRED a FREDi.                                        *
 * Pet funkcnich tlacitek F0 - F4 ovlada prislusne funkce. V kombinaci se stiskem tlacitka Shift   *
 * pak funkce F1 - F4 -> F5 - F8. Tlacitko EmStop okamzite zastavi lokomotivu, odblokovani tohoto  *                                        
 * stavu se provede stiskem Shift a F0. Ovladac si pamatuje stisknuti tlacitka Shift, tlacitka     *
 * neni treba drzet obe naraz. Ovladac neumoznuje vlastni volbu adresy (stejne jako FRED), umi     *
 * pouze prijmout adresu vyslanou centralou nebo jinym ovladacem, ktery umi funkci Dispatch.       *
 * Po prepnuti prepinace je adresa opet uvolnena a muze byt prijata jinym ovladacem. Adresa a pri- *
 * razeny slot jsou ulozeny v EEPROM pameti ovladace, takze jej lze odpojit od site a pripojit na  *
 * jinem miste a pokracovat v ovladani puvodni lokomotivy. Prirazeni adresy ovladaci je indikovano * 
 * modrou LED - bez adresy blika, s adresou sviti trvale. Rychlost jizdy je ovladana potenciomet-  *
 * rem. Maximalni rychlost je vyhodnocena jako chybovy stav a ovladac okamzite zastavi lokomotivu  *
 * (stejne jako pri EmStop). Duvodem je casty jev (zejmena u deti) nechteneho prudkeho pretoceni   *
 * potenciometru na nespravnou stranu ve snaze rychle zastavit lokomotivu. Uvolni se pootocenim    *
 * zpet a stiskem Shift - F0. Trojpolohovy prepinac On-Off-On voli smer jizdy. Zvoleny smer je in- *
 * dikovan zlutou (vpred) a cervenou (vzad) LED. Posledni, zelena dioda, indikuje provoz v Loco-   *
 * Netu. Software ovladace NIJAK NERESI mozne kolizni stavy typu adresa ci slot byly behem jeho    *
 * odpojeni prideleny jinemu ovladaci.                                                             *
 * Zakmity pri stisku tlacitka resi knihovna Button, ukladani vicebytovych adres do EEPROM pak     *
 * knihovna EEPROMEx http://thijs.elenbaas.net/2012/07/extended-eeprom-library-for-arduino/        *
 *                                                                                                 *
 * Software je volne siritelny, modifikovatelny, znicitelny. Pripadne pouziti pouze na vlastni     *
 * riziko!!                                                                                        *
 *                                                                                                 *
 * Hardware - zakladem ovladace je Arduino Pro Mini. Komunikace mezi procesorem a sbernici LocoNet *                                                                                
 * je zajistena prevodnikem podle Philippa Gahtowa http://pgahtow.de/wiki/index.php?title=Loconet  *
 * s drobnou upravou podle Jindry Fucika http://fucik.name/Arduino/begin.php                       *
 * Napajeni je realizovano usmernenym napetim z RailSync +/-                                       *
 *                                                                                                 *
 * 25.05.2016                                                                   JP                 *                                                                                               
 *                                                                                                 *
 ***************************************************************************************************/
#include <LocoNet.h>
#include <EEPROMex.h>
#include <Button.h>

// Prirazeni pinu

#define LOCONET_TX_PIN 9
#define PotPin A1  		        // Vstup potenciometru
#define DispatchSwitch 14  	  // Tlacitko pro prijeti/uvolneni adresy lokomotivy
#define F0Button 2  		      // Tlacitko F0
#define F1Button 5            // Tlacitko F1
#define F2Button 7            // Tlacitko F2
#define F3Button 4            // Tlacitko F3
#define F4Button 6            // Tlacitko F4
#define ShiftButton 3         // Tlacitko Shift
#define EmStopButton 18       // Tlacitko Emergency Stop
#define DirSwitchF 11 		    // Prepinac smeru, dopredu (Forward)
#define DirSwitchR 10  		    // Prepinac smeru, dozadu (Reverse)
#define RedLED 17 			      // Indikace smeru (R)
#define YellowLED 12          // Indikace smeru (F)
#define GreenLED 13  		      // Indikace provozu
#define StatusLED 16  	      // Indikator prijeti adresy
#define EEPROMAdr 200         // Pocatecni adresa pro ulozen adresy loko
#define EEPROMSlot 20         // Adresa pro ulozeni slotu

// Ostatni promenne

int sensorValue = 0;  			      // Hodnoty ctene z potenciometru
int outputValue = 0;  			      // Hodnoty premapovane z potenciometru
int Adresa;  				              // Adresa lokomotivy
uint8_t Slot = 0; 			          // Cislo slotu zarizeni
byte TickCount = 0;  			        // Pocitadlo blikani StatusLED
long blik = 400; 			            // Interval mezi zmenou stavu StatusLED
unsigned long predcas = 0;		    // Promenna pro zmenu stavu StatusLED
int Redstatus = 0;  			        // Stav cervene diody
int Yellowstatus = 0;  			      // Stav zlute diody
int Smer = 0; 				            // Kontrola prepnuti smeru; 1 - dopredu; 0 - dozadu; 2 - neutral; 3 - emergency stop
LocoNetThrottleClass Throttle ;
lnMsg *RxPacket ;
uint32_t LastThrottleTimerTick;
bool SH = 0;

// Tlacitka

  Button F0(F0Button);
  Button F1(F1Button);
  Button F2(F2Button);
  Button F3(F3Button);
  Button F4(F4Button);
  Button Shift(ShiftButton);
  Button EmStop(EmStopButton);
  
// Setup

void setup() {
  pinMode(DispatchSwitch, INPUT_PULLUP);
  pinMode(DirSwitchF, INPUT_PULLUP);
  pinMode(DirSwitchR, INPUT_PULLUP);
  pinMode(YellowLED, OUTPUT);
  pinMode(RedLED, OUTPUT);
  pinMode(StatusLED, OUTPUT);
  F0.begin();
  F1.begin();
  F2.begin();
  F3.begin();
  F4.begin();
  Shift.begin();
  EmStop.begin();
 
  LocoNet.init(LOCONET_TX_PIN); 	// Inicializace LocoNetu na TX_PIN

  Throttle.init(0, 0, 9999);  		// Inicializace ovladace na adrese 9999
  
  getAddress(1);  			          //  Pri vypadku napajeni pouzije aresu/slot z EEPROM

}

boolean isTime(unsigned long *timeMark, unsigned long timeInterval)
{
    unsigned long timeNow = millis();
    if ( timeNow - *timeMark >= timeInterval)
    {
        *timeMark = timeNow;
        return true;
    }    
    return false;   
}

// Hlavni smycka

void loop() { 
  RxPacket = LocoNet.receive() ; 	// Precist LocoNet paket
  
  if (RxPacket) 			
  {    
    //digitalWrite(13, HIGH);
    
    if(!LocoNet.processSwitchSensorMessage(RxPacket))
      Throttle.processMessage(RxPacket) ;  
  }
  
  if (isTime(&LastThrottleTimerTick, 100))
  {
    Throttle.process100msActions(); 
    //digitalWrite(13, LOW);
  }

  getAddress(0);  			// Nacti adresu lokomotivy

  getDirection();  			// Zjisti smer

  getFunction();  			// Funkce

  TickLED();  				  // Indikacni dioda

  getSpeed();  				  // Rychlost jizdy

  digitalWrite(YellowLED, Yellowstatus);  // Rozsvit / zhasni zlutou LED
  digitalWrite(RedLED, Redstatus); 	      // Rozsvit / zhasni cervenou LED
  digitalWrite(GreenLED, SH);             // Rozsvit / zhasni zelenou diodu po namacknuti Shift
}

// Subrutiny

void TickLED ()  			// Pri neobsazene adrese lokomotivy blika, pri nacteni sviti trvale
{
  if (Throttle.getState() == TH_ST_IN_USE)
    {
      digitalWrite(StatusLED, HIGH);   
      return;      
    }
  
  if (Throttle.getState() != TH_ST_IN_USE) 
    {
      unsigned long cas = millis();
      if (cas - predcas > blik)
      {
        predcas = cas;
        digitalWrite(StatusLED, !digitalRead(StatusLED));  
      }
      return;
    }
}

//----

void getSpeed()   			// Pokud je smer nastaven na 0 ci 1, rozjede lokomotivu podle pozice potenciometru, v mezipoloze zastavi podle programu, pri EmStop ihned
{
  if (Throttle.getState() == TH_ST_IN_USE)
  {
    if (Smer != 3)
      {
        if (Smer != 2)
        {
          sensorValue = analogRead(PotPin);
          outputValue = map(sensorValue, 0, 1020, 127, 1);
          Throttle.setSpeed(outputValue);
          if (outputValue == 127)
          {
            Smer = 3;       // Pojistka proti nechtenemu pretoceni ovladace rychlosti na max.
          }
        }
      }
    
  if (Smer == 2)
    { 
      Throttle.setSpeed(1); 
    }
    
  if (Smer == 3)
    {
      Throttle.setSpeed(0);
    }
  return;
  }
}

//----

void getDirection()  			// Pri prepnuti prepinace zmeni nastaveni smeru a rozsviti prislusne kontrolni LED. V poloze 'off' zastavi lokomotivu
                                          // Pri emergency stop na zmenu prepinace nereaguje az do stisknuti Shift + F0
{
  if (Throttle.getState() == TH_ST_IN_USE)
  {
    if (digitalRead(DirSwitchF) == LOW && Smer != 3)
    {
      if (Smer != 1)
      {
        Throttle.setDirection(1);
        Redstatus = 1;
        Yellowstatus = 0;        
        Smer = 1;    
        return;
      }
    }
     
    if (digitalRead(DirSwitchR) == LOW && Smer != 3)
    {
      if (Smer != 0)
      {
        Throttle.setDirection(0);
        Redstatus = 0;
        Yellowstatus = 1;
        Smer = 0;
        return;
      }
    }
    
    if (digitalRead(DirSwitchR) == HIGH && digitalRead(DirSwitchF) == HIGH && Smer != 3)  // Prepinac v nulove poloze zastavi vlak
    { 
    if (Smer != 2)
      {       
        Redstatus = 0;
        Yellowstatus = 0;
        Smer = 2;
        return;
      }
    }
  }
}

//-----

void getAddress(uint8_t Restart)  // Pri startu ovladace nastavi posledni znamou adresu a slot. Pri normalnim provozu priradi / uvolni adresu
{
  if (Restart)
  {
    if (digitalRead(DispatchSwitch) == LOW)
    {
      Throttle.resumeAddress(EEPROM.readInt(EEPROMAdr), EEPROM.read(EEPROMSlot));
      
      if (digitalRead(DirSwitchF) == LOW)
      {   
        Smer = 0;    
        return;
      }
        
      if (digitalRead(DirSwitchR) == LOW)
      {
        Smer = 1;
        return;
      }
    
      if (digitalRead(DirSwitchR) == HIGH && digitalRead(DirSwitchF) == HIGH)  
      {      
        Smer = 4;
        return;     
      }
      return;
    }
    
    if (digitalRead(DispatchSwitch) == HIGH)
    {
      return;
    }
  }
  
  if (digitalRead(DispatchSwitch) == HIGH)
  {
    if (Throttle.getState() == TH_ST_IN_USE)
    {
      Throttle.releaseAddress(); 
      Throttle.dispatchAddress(Adresa); 
      return;
    }
  }
  
  if (digitalRead(DispatchSwitch) == LOW)
  { 
    if (Throttle.getState() != TH_ST_IN_USE)
    {
      switch (Throttle.getState())
      {
        case TH_ST_FREE:
            Throttle.setAddress(Throttle.acquireAddress()); 
        break;
      }
    }
  }
}

//-----

void getFunction()  		// Pri stisku tlacitka zapne/vypne F0 - F4, Shift + F1 - F4 -> F5 - F8, EmStop - emergency stop, Shift + F0 - normalni rezim
{
  if (!SH)
  {
    if (F0.pressed())  
    {
      Throttle.setFunction(0, !Throttle.getFunction(0));
      return;
    }

    if (F1.pressed())
    {
      Throttle.setFunction(1, !Throttle.getFunction(1));
      return;
    }
  
    if (F2.pressed())
    {
      Throttle.setFunction(2, !Throttle.getFunction(2));
      return;
    }
  
    if (F3.pressed())
    {
      Throttle.setFunction(3, !Throttle.getFunction(3));
      return;
    }
  
    if (F4.pressed())
    {
      Throttle.setFunction(4, !Throttle.getFunction(4));
      return;
    }
  
    if (Shift.pressed())
    {
      SH = 1;
    }
  }
  
  if (SH)
  {
    if (F1.pressed())  
    {
      Throttle.setFunction(5, !Throttle.getFunction(5));
      SH = 0;
      return;
    }
  
    if (F2.pressed())
    {
      Throttle.setFunction(6, !Throttle.getFunction(6));
      SH = 0;
      return;
    }
  
    if (F3.pressed())
    {
      Throttle.setFunction(7, !Throttle.getFunction(7));
      SH = 0;
      return;
    }
  
    if (F4.pressed())
    {
      Throttle.setFunction(8, !Throttle.getFunction(8));
      SH = 0;
      return;
    }  
  
    if (F0.pressed() && Smer == 3)
    {
      SH = 0;
    Smer = 4;
    return;
    }
    
    if (Shift.pressed())          // Pokud doslo k namacknuti tlacitka Shift omylem, opetovnym stiskem se zrusi.
    {
      SH = 0;  
      return;
    }
  }
  
  if (EmStop.pressed())
  {
   Smer = 3;
   return;
  } 
}

//-----

void notifyThrottleAddress( uint8_t UserData, TH_STATE State, uint16_t Address, uint8_t isSlot )
{
  if (Throttle.getState() == TH_ST_IN_USE)
  {
    Adresa = Address;
    Slot = isSlot;
    EEPROM.updateInt(EEPROMAdr, Adresa);  	// Ulozi adresu lokomotivy do EEPROM
    EEPROM.update(EEPROMSlot, Slot);		    // Ulozi slot do EEPROM
  }
}








